#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <omp.h>

#include "genlib.h"
#include "PPPostProcess.h"
#include "PhyCorrect.h"
#include "KernelDensity.h"

//#define HTML_OUTPUT


#ifdef _WIN32
	#include <direct.h>

	void	mdir(char *DirName)
	{
		_mkdir(DirName);
	}

#else
	#include<sys/stat.h>

	void	mdir(char *DirName)
	{
		mkdir(DirName, 0777);
	}
#endif

void	FinalizeSPrams(DATA *Data);

TAXA*	CreatTaxa(char *No, char *Name)
{
	TAXA *Ret;

	Ret = (TAXA*)malloc(sizeof(TAXA));
	if(Ret == NULL)
		MallocErr();

	Ret->No		= atoi(No);
	Ret->Name	= StrMake(Name);
	Ret->MeanRT = 0;
	Ret->SDRT	= 0;

	return Ret;
}

SPRAM*	AddSP(NODE Node)
{
	SPRAM* Ret;

	Ret = (SPRAM*)malloc(sizeof(SPRAM));
	if(Ret == NULL)
		MallocErr();

	Ret->Node		= Node;

	Ret->MeanBL		= 0;
	Ret->MeanNode	= 0;

	Ret->NoBL		= 0;
	Ret->NoNode		= 0;

	Ret->SDBL		= 0;
	Ret->SDNode		= 0;

	return Ret;
}

NODE	MakeNode(int Tokes, char **Passed)
{
	NODE N;
	int	 Index;

	N = (NODE)malloc(sizeof(struct INODE));
	if(N == NULL)
		MallocErr();

	N->BLList = NULL;
	N->ID = atoi(Passed[0]);
	N->Len = atof(Passed[1]);
	N->TLen = N->Len;

	N->NoPart = atoi(Passed[2]);

	N->Part = (int*)malloc(sizeof(int) * N->NoPart);
	if(N->Part == NULL)
		MallocErr();

	for(Index=0;Index<N->NoPart;Index++)
		N->Part[Index] = atoi(Passed[Index+3]);

	N->NoNodes = 0;
	N->NList = NULL;
	N->Ans = NULL;

	if(N->NoPart == 1)
		N->Tip = TRUE;
	else
		N->Tip = FALSE;

	N->Scale = 1;
	N->Pos = -1;

	N->SPram = AddSP(N);

	N->Height = -1;


	N->KDSigV = NULL;
	N->KDDeltaB = NULL;
	N->KDDeltaV = NULL;
	N->KDR = NULL;

	return N;
}

BLINFO*	AllocBLInfo(DATA *Data)
{
	BLINFO *BLInfo;
	int		Index;

	BLInfo = (BLINFO*)malloc(sizeof(BLINFO));
	if(BLInfo == NULL)
		MallocErr();

	BLInfo->BL = (double*)malloc(sizeof(double) * Data->NoNodes);
	if(BLInfo->BL == NULL) MallocErr();

	BLInfo->Mean = (double*)malloc(sizeof(double) * Data->NoNodes);
	if(BLInfo->Mean == NULL) MallocErr();

	BLInfo->SD = (double*)malloc(sizeof(double) * Data->NoNodes);
	if(BLInfo->SD == NULL) MallocErr();

	BLInfo->NoDiff = (int*)malloc(sizeof(int) * Data->NoNodes);
	if(BLInfo->NoDiff == NULL) MallocErr();

	BLInfo->MeanScale = (double*)malloc(sizeof(double) * Data->NoNodes);
	if(BLInfo->MeanScale == NULL) MallocErr();

	for(Index=0;Index<Data->NoNodes;Index++)
	{
		BLInfo->BL[Index] = Data->NodeList[Index]->Len;
		BLInfo->Mean[Index] =  0;
		BLInfo->SD[Index] = 0;
		BLInfo->NoDiff[Index] = 0;
		BLInfo->MeanScale[Index] = 0;
	}

	BLInfo->No = 0;

	return BLInfo;
}

void	FreeBLInfo(DATA *Data)
{
	BLINFO *BLInfo;

	BLInfo = Data->BLInfo;

	free(BLInfo->BL);
	free(BLInfo->Mean);
	free(BLInfo->SD);
	free(BLInfo->MeanScale);
	free(BLInfo->NoDiff);

	free(BLInfo);
}

void	PrintBLInfo(DATA *Data)
{
	int Index;
	BLINFO *BLInfo;

	BLInfo = Data->BLInfo;

	for(Index=0;Index<Data->NoNodes;Index++)
	{
		printf("%d\t%f\t", Index, BLInfo->BL[Index]);
		printf("%f\t%f\n", BLInfo->Mean[Index], BLInfo->MeanScale[Index]);
	}
}

void	FreeRates(RATES *Rates)
{
	free(Rates->Name);
	free(Rates->Rate);
	free(Rates->Cats);
	free(Rates->NoBL);
	free(Rates->MeanRates);
	free(Rates->SDRates);

	free(Rates);
}

void	FreeNode(NODE N)
{
	if(N->SPram != NULL)
		free(N->SPram);

	if(N->BLList != NULL)
		free(N->BLList);

	FreeKD(N->KDR);
	FreeKD(N->KDDeltaB);
	FreeKD(N->KDDeltaV);
	FreeKD(N->KDSigV);

	free(N->Part);
	free(N->NList);
	free(N);
}

void	FreeTaxa(TAXA *T)
{
	free(T->Name);
	free(T);
}

int		GetTaxaID(char *Taxa, DATA *Data)
{
	int Index;

	for(Index=0;Index<Data->NoTaxa;Index++)
	{
		if(strcmp(Taxa, Data->TaxaList[Index]->Name) == 0)
			return Data->TaxaList[Index]->No;
	}

	return -1;
}

int		GetNoNodes(NODE N)
{
	int	No;

	No = 0;

	while(N != NULL)
	{
		No++;
		N = N->Ans;
	}

	return No;
}

double	SumScaleToRoot(DATA *Data, NODE N)
{
	double Ret;
	BLINFO *BLInfo;

	BLInfo = Data->BLInfo;

	Ret = 0;
	while(N != NULL)
	{
		Ret += BLInfo->MeanScale[N->Pos];
//		Ret += N->Scale;
		N = N->Ans;
	}

	return Ret;
}

double	GetNodeToRoot(NODE N)
{
	double Dist;

	Dist = 0;

	while(N != NULL)
	{
		Dist += N->Len;
		N = N->Ans;
	}

	return Dist;
}

double	MaxRootToTip(DATA *Data)
{
	int		Index;
	NODE	N;
	double	Max;
	double	Dist;

	Data->Root->Len = 0;
	Data->Root->TLen = 0;

	Max = 0;
	for(Index=0;Index<Data->NoNodes;Index++)
	{
		N = Data->NodeList[Index];
		if(N->Tip == TRUE)
		{
			Dist = GetNodeToRoot(N);
			if(Dist > Max)
				Max = Dist;
		}
	}

	return Max;
}

void	SetRatesCats(RATES *Rates)
{
	int		Index;
	double	Temp, Size;

	Rates->Cats = (double*)malloc(sizeof(double) * Rates->NoCats);
	if(Rates->Cats == NULL)
		MallocErr();

	Size = Rates->TL / Rates->NoCats;

	for(Index=0;Index<Rates->NoCats;Index++)
	{
//		Rates->Cats[Index] = Size * (1+Index);
		Temp = ((Index+1)-0.5) / (double)Rates->NoCats;
		Rates->Cats[Index] = Temp * Rates->TL;
//		printf("%d\t%f\n", Index, Rates->Cats[Index]);
	}

//	exit(0);
}

void	SetRateCat(DATA *Data, RATES *Rates)
{
	int		Index;
	OPTIONS	*Opt;

	Opt = Data->Opt;
	Rates->NoCats = Opt->NoCats;

	SetRatesCats(Rates);

	Rates->Rate = (double*)malloc(sizeof(double) * Data->NoTaxa);
	if(Rates->Rate == NULL) MallocErr();

	Rates->NoBL = (int*)malloc(sizeof(int)*  Rates->NoCats);
	if(Rates->NoBL == NULL) MallocErr();

	Rates->MeanRates = (double*)malloc(sizeof(double) * Rates->NoCats);
	if(Rates->MeanRates == NULL) MallocErr();

	Rates->SDRates = (double*)malloc(sizeof(double) * Rates->NoCats);
	if(Rates->SDRates == NULL) MallocErr();

	for(Index=0;Index<Data->NoTaxa;Index++)
		Rates->Rate[Index]		= 0.0;

	for(Index=0;Index<Rates->NoCats;Index++)
	{
		Rates->MeanRates[Index] = 0.0;
		Rates->SDRates[Index]	= 0.0;
		Rates->NoBL[Index]		= 0;
	}
}

RATES*	MakeRate(char *Name, NODE Node, DATA *Data)
{
	RATES *Rates;

	Rates = (RATES*)malloc(sizeof(RATES));
	if(Rates == NULL) MallocErr();

	Rates->Name = StrMake(Name);
	Rates->StartNode = Node;

	Rates->StatPos = GetNodeToRoot(Node);

	if(Node->Ans != NULL)
		Rates->StatPos = Rates->StatPos - Node->TLen;

	Rates->TL = MaxRootToTip(Data);

	Rates->Rate = NULL;
	Rates->NoBL = NULL;
	Rates->MeanRates = NULL;
	Rates->SDRates = NULL;

	Rates->Cats = NULL;
	Rates->NoCats = 0;

	return Rates;
}

int		TaxaInNode(int TaxaID, NODE N)
{
	int Index;

	for(Index=0;Index<N->NoPart;Index++)
		if(N->Part[Index] == TaxaID)
			return TRUE;

	return FALSE;
}

int		ValidRateNode(NODE N, int *List)
{
	int Index;

	for(Index=0;Index<N->NoPart;Index++)
	{
		if(TaxaInNode(List[Index], N) == FALSE)
			return FALSE;
	}

	return TRUE;
}

NODE	GetNodeFromList(DATA* Data, int Size, int* List)
{
	int Index;
	NODE Node;

	for(Index=0;Index<Data->NoNodes;Index++)
	{
		Node = Data->NodeList[Index];
		if(Node->NoPart == Size)
		{
			if(ValidRateNode(Node, List) == TRUE)
				return Node;
		}
	}

	return NULL;
}

NODE	GetNodeFromTaxaList(DATA* Data, int Size, char **TList)
{
	NODE	Ret;
	int		Index;
	int		*List;

	List = (int*)malloc(sizeof(int) * Size);
	if(List == NULL)
		MallocErr();

	for(Index=0;Index<Size;Index++)
		List[Index] = GetTaxaID(TList[Index], Data);

	Ret = GetNodeFromList(Data, Size, List);

	free(List);

	return Ret;
}


void	MakeRateFormLine(DATA *Data, int Tokes, char **Passed)
{
	int		No;
	int		*PList;
	int		Index;
	char	*Name;
	NODE	Node;
	RATES	*Rates;

	No = Tokes - 1;
	PList = (int*)malloc(sizeof(int) * No);
	if(PList == NULL) MallocErr();

	Name = Passed[0];

	for(Index=1;Index<Tokes;Index++)
	{
		PList[Index-1] = GetTaxaID(Passed[Index], Data);

		if(PList[Index-1] == -1)
		{
			printf("Cannot find matching taxa for %s node %s\n", Passed[Index], Name);
			exit(0);
		}
	}

	Node = 	GetNodeFromList(Data, No, PList);

	if(Node == NULL)
	{
		printf("Count not find valid node for %s\n", Name);
		exit(0);
	}

	Rates = MakeRate(Name, Node, Data);

	Data->Opt->Rates = (RATES**)AddToList(&Data->Opt->NoRates, (void**)Data->Opt->Rates, (void*)Rates);

	free(PList);
}

void	AddUserNodeRatio(DATA *Data, int NoTokes, char **Passed)
{
	OPTIONS		*Opt;
	NODERATIO	*NodeRatio;
	int			Index;

	Opt = Data->Opt;

	NodeRatio = (NODERATIO*)malloc(sizeof(NODERATIO));
	if(NodeRatio == NULL)
		MallocErr();

	NodeRatio->Name = StrMake(Passed[0]);
	NodeRatio->NoTaxa = NoTokes - 1;

	NodeRatio->Node = NULL;

	NodeRatio->TNames = (char**)malloc(sizeof(char*) * NodeRatio->NoTaxa);
	if(NodeRatio->TNames == NULL)
		MallocErr();

	for(Index=0;Index<NodeRatio->NoTaxa;Index++)
		NodeRatio->TNames[Index] = StrMake(Passed[Index+1]);

	Opt->NodeRatio = (NODERATIO**)AddToList(&Opt->NoNodeRatio, (void**)Opt->NodeRatio, (void*)NodeRatio);
}

void	MakeSlice(DATA *Data, char *Name, double Point)
{
	SLICE	*Slice;
	int		Index;
	OPTIONS	*Opt;

	Slice = (SLICE*)malloc(sizeof(SLICE));
	if(Slice == NULL)
		MallocErr();

	Slice->Point = Point;
	Slice->Name = StrMake(Name);

	Slice->Mean = (double*)malloc(sizeof(double) * Data->NoTaxa);
	if(Slice->Mean == NULL)
		MallocErr();

	Slice->SD = (double*)malloc(sizeof(double) * Data->NoTaxa);
	if(Slice->SD == NULL)
		MallocErr();

	Slice->NodeNo = (int*)malloc(sizeof(int) * Data->NoTaxa);
	if(Slice->NodeNo == NULL)
		MallocErr();

	for(Index=0;Index<Data->NoTaxa;Index++)
	{
		Slice->Mean[Index] = 0;
		Slice->SD[Index] = 0;
		Slice->NodeNo[Index] = -1;
	}

	Opt = Data->Opt;

	Opt->Slices = (SLICE**)AddToList(&Opt->NoSlices, (void**)Opt->Slices, (void*)Slice);
}

void	FreeSlice(DATA *Data)
{
	int		Index;
	OPTIONS *Opt;
	SLICE	*Slice;

	Opt = Data->Opt;

	if(Opt->NoSlices == 0)
		return;

	for(Index=0;Index<Opt->NoSlices;Index++)
	{
		Slice = Opt->Slices[Index];

		free(Slice->Name);
		free(Slice->Mean);
		free(Slice->SD);
		free(Slice->NodeNo);
		free(Slice);
	}

	free(Opt->Slices);

	Opt->Slices = NULL;
	Opt->NoSlices = -1;
}

void SetNodeRatio(DATA *Data, double Ratio, int MinRatioSize)
{
	OPTIONS *Opt;

	Opt = Data->Opt;

	Opt->FindRatio = TRUE;
	Opt->MinRatioSize = MinRatioSize;
	Opt->Ratio = Ratio;
}

OPTIONS*	CreatOptions(void)
{
	OPTIONS		*Ret;

	Ret = (OPTIONS*)malloc(sizeof(OPTIONS));
	if(Ret == NULL)
		MallocErr();

	Ret->StartIt	= 0;
	Ret->NoCats		= 1000;

	Ret->NoRates	= 0;
	Ret->Rates		= NULL;
	
	Ret->NoSlices	= 0;
	Ret->Slices		= NULL;
		
	Ret->NoNodeRatio = 0;
	Ret->NodeRatio	= NULL;
	
	Ret->Normalise	= FALSE;
	
	Ret->OptFN		= NULL;

	Ret->FindRatio	= FALSE;
	
	return Ret;
}

void	LoadOptions(DATA *Data, char *FName)
{
	OPTIONS		*Opt;
	char		*Buffer, **Passed;
	TEXTFILE	*TF;
	int			Index, Tokes;

	Opt = Data->Opt;
		
	Opt->OptFN = StrMake(FName);
		
	Opt->Rates = (RATES**)malloc(sizeof(RATES*));
	if(Opt->Rates == NULL)
		MallocErr();
	Opt->Rates[0] = MakeRate("Root", Data->Root, Data);
	Opt->NoRates = 1;

	TF = LoadTextFile(Opt->OptFN, FALSE);

	Buffer = (char*)malloc(sizeof(char) * (TF->MaxLine + 1));
	Passed = (char**)malloc(sizeof(char*) * (TF->MaxLine + 1));
	if((Buffer == NULL) || (Passed == NULL))
		MallocErr();

	for(Index=0;Index<TF->NoOfLines;Index++)
	{
		strcpy(Buffer, TF->Data[Index]);
		Tokes = MakeArgv(Buffer, Passed, TF->MaxLine);
		if(Tokes >= 1)
		{
			MakeUpper(Passed[0]);

			if(strcmp("IT", Passed[0]) == 0)
				Opt->StartIt = atoi(Passed[1]);

			if(strcmp("NOCATS", Passed[0]) == 0)
				Opt->NoCats = atoi(Passed[1]);

			if(strcmp("RATES", Passed[0]) == 0)
				MakeRateFormLine(Data, Tokes-1, &Passed[1]);

			if(strcmp("SLICE", Passed[0]) == 0)
				MakeSlice(Data, Passed[1], atof(Passed[2]));

			if(strcmp("NODERATIO", Passed[0]) == 0)
				SetNodeRatio(Data, atof(Passed[1]), atoi(Passed[2]));

			if(strcmp("RATIONODE", Passed[0]) == 0)
				AddUserNodeRatio(Data, Tokes-1, &Passed[1]);

			if(strcmp("NORMALISE", Passed[0]) == 0)
				Opt->Normalise = TRUE;
		}
	}

	for(Index=0;Index<Opt->NoRates;Index++)
		SetRateCat(Data, Opt->Rates[Index]);

	free(Buffer);
	free(Passed);

	FreeTextFile(TF);
}

void	FreeData(DATA *Data)
{
	int	Index;

	FreeSlice(Data);
	FreeBLInfo(Data);

	for(Index=0;Index<Data->Opt->NoRates;Index++)
		FreeRates(Data->Opt->Rates[Index]);
	free(Data->Opt->Rates);
	free(Data->Opt->OptFN);
	free(Data->Opt);

	for(Index=0;Index<Data->NoNodes;Index++)
		FreeNode(Data->NodeList[Index]);
	free(Data->NodeList);

	for(Index=0;Index<Data->NoTaxa;Index++)
		FreeTaxa(Data->TaxaList[Index]);
	free(Data->TaxaList);

	free(Data->NodeOrder);

	free(Data->FName);
	free(Data->Sig2);
	free(Data);
}

DATA*	LoadData(char *FName)
{
	DATA	*Data;
	char	*Buffer;
	int		Index;
	char	**Passed;
	int		Tokes;

	Data = (DATA*)malloc(sizeof(DATA));
	if(Data == NULL) MallocErr();

	Data->FName = StrMake(FName);

	Buffer = (char*)malloc(sizeof(char) * BUFFERSIZE);
	Passed = (char**)malloc(sizeof(char*) * BUFFERSIZE);
	if((Buffer == NULL) || (Passed == NULL)) MallocErr();

	Data->DFile = OpenRead(Data->FName);

	fgets(Buffer, BUFFERSIZE, Data->DFile);

	Data->NoTaxa = atoi(Buffer);
	if(Data->NoTaxa < 1)
	{
		printf("%s needs more that one node.\n", Data->FName);
		exit(0);
	}

	Data->TaxaList = (TAXA**)malloc(sizeof(TAXA*) * Data->NoTaxa);
	if(Data->TaxaList == NULL) MallocErr();
	for(Index=0;Index<Data->NoTaxa;Index++)
	{
		fgets(Buffer, BUFFERSIZE, Data->DFile);
		Tokes = MakeArgv(Buffer, Passed, BUFFERSIZE);
		if(Tokes != 2)
		{
			printf("Err finding taxa:\n");
			exit(1);
		}

		Data->TaxaList[Index] = CreatTaxa(Passed[0], Passed[1]);
	}

	fgets(Buffer, BUFFERSIZE, Data->DFile);

	Data->NoNodes = atoi(Buffer);
	if(Data->NoTaxa < 1)
	{
		printf("%s needs more that one node.\n", Data->FName);
		exit(0);
	}

	Data->NodeList = (NODE*)malloc(sizeof(NODE) * Data->NoNodes);
	if(Data->NodeList == NULL)
		MallocErr();
	for(Index=0;Index<Data->NoNodes;Index++)
	{
		fgets(Buffer, BUFFERSIZE, Data->DFile);
		Tokes = MakeArgv(Buffer, Passed, BUFFERSIZE);
		Data->NodeList[Index] = MakeNode(Tokes, Passed);
	}

	free(Buffer);
	free(Passed);

	Data->NodeOrder = (int*)malloc(sizeof(int) * Data->NoNodes);
	if(Data->NodeOrder == NULL)
		MallocErr();
	for(Index=0;Index<Data->NoNodes;Index++)
		Data->NodeOrder[Index] = Data->NodeList[Index]->ID;

	Data->BLInfo = AllocBLInfo(Data);


	Data->TreeNo	= 0;
	Data->NoTrees	= 0;

	Data->Sig2		= NULL;

	return Data;
}

int		FindRoot(DATA *Data)
{
	int No;
	int Index;

	No = 0;
	for(Index=0;Index<Data->NoNodes;Index++)
	{
		if(Data->NodeList[Index]->NoPart == Data->NoTaxa)
			return Index;

	}

	printf("Error cannot find root.\n");
	exit(0);
	return -1;
}

int		FindStart(DATA *Data)
{
	int Index;

	for(Index=0;Index<Data->NoNodes;Index++)
	{
		if(Data->NodeList[Index] != NULL)
			return Index;
	}

	return -1;
}

int		FindNextPart(DATA *Data)
{
	int No, Index;

	No = FindStart(Data);
	if(No == -1)
		return -1;

	for(Index=No+1;Index<Data->NoNodes;Index++)
	{
		if(Data->NodeList[Index] != NULL)
		{
			if(Data->NodeList[Index]->NoPart > Data->NodeList[No]->NoPart)
				No = Index;
		}
	}

	return No;
}





/* Is A a subset of B? */
int		SubSet(NODE A, NODE B)
{
	int	Index;

	if(A->NoPart > B->NoPart)
		return FALSE;

	for(Index=0;Index<A->NoPart;Index++)
	{
		if(TaxaInNode(A->Part[Index], B) == FALSE)
			return FALSE;
	}

	return TRUE;
}
/*
int		PlaceNewNode(NODE NewN, NODE N)
{
	int Placed, Index;

	if(N->Tip == TRUE)
		return FALSE;

	if(SubSet(NewN, N) == FALSE)
		return FALSE;

	for(Index=0;Index<N->NoNodes;Index++)
	{
		Placed = PlaceNewNode(NewN, N->NList[Index]);
		if(Placed == TRUE)
			return TRUE;
	}

	NewN->Ans = N;
	N->NList = (NODE*)AddToList(&N->NoNodes, (void**)N->NList, (void*)NewN);

	return TRUE;
}
*/

int		PlaceNewNode(NODE NewN, NODE N)
{
	int Placed, Index;

	if(N->Tip == TRUE)
		return FALSE;

	if(SubSet(NewN, N) == FALSE)
		return FALSE;

	for(Index=0;Index<N->NoNodes;Index++)
	{
		Placed = PlaceNewNode(NewN, N->NList[Index]);
		if(Placed == TRUE)
			return TRUE;
	}

	NewN->Ans = N;
	N->NList = (NODE*)AddToList(&N->NoNodes, (void**)N->NList, (void*)NewN);

	return TRUE;
}

NODE	FindNode(NODE N, int ID)
{
	int Index;
	NODE NRet;

	if(N->ID == ID)
		return N;

	for(Index=0;Index<N->NoNodes;Index++)
	{
		NRet = FindNode(N->NList[Index], ID);
		if(NRet != NULL)
			return NRet;
	}

	return NULL;
}

NODE	GetNode(DATA *Data, int ID)
{
	int Index;

	for(Index=0;Index<Data->NoNodes;Index++)
		if(Data->NodeList[Index]->ID == ID)
			return Data->NodeList[Index];

	return NULL;
}

void	FlattenNodes(DATA *Data)
{
	int Index;

	for(Index=0;Index<Data->NoNodes;Index++)
	{
		Data->NodeList[Index] = FindNode(Data->Root, Data->NodeOrder[Index]);
		Data->NodeList[Index]->Pos = Index;
	}
}

TAXA*	GetTaxa(int ID, DATA *Data)
{
	int Index;

	for(Index=0;Index<Data->NoTaxa;Index++)
		if(Data->TaxaList[Index]->No == ID)
			return Data->TaxaList[Index];

	return NULL;
}

void	SetTips(DATA *Data)
{
	int Index;
	NODE N;

	for(Index=0;Index<Data->NoNodes;Index++)
	{
		N = Data->NodeList[Index];

		if(N->Tip == TRUE)
			N->Taxa = GetTaxa(N->Part[0], Data);
	}
}


void	BuildTree(DATA *Data)
{
	int RootID;
	int	NextNo;
	NODE N;

 	RootID = FindRoot(Data);

	Data->Root = Data->NodeList[RootID];
	Data->NodeList[RootID] = NULL;

	do
	{
		NextNo = FindNextPart(Data);
		if(NextNo != -1)
		{
			N = Data->NodeList[NextNo];
			PlaceNewNode(N, Data->Root);
			Data->NodeList[NextNo] = NULL;
		}
	} while(NextNo != -1);


	Data->Root->Ans = NULL;
	Data->BLInfo->BL[RootID] = 0;
	Data->Root->Len = 0;

	FlattenNodes(Data);
	SetTips(Data);

	Data->MaxRootToTip = MaxRootToTip(Data);

/*	RootToTip(Data);
	SetRateCat(Data);
*/
}

void	ReSetTree(DATA *Data)
{
	int Index;

	for(Index=0;Index<Data->NoNodes;Index++)
	{
		Data->NodeList[Index]->Len = Data->BLInfo->BL[Index];
		Data->NodeList[Index]->Scale = 1;
	}
}

void	ScaleNode(NODE N, double Scale)
{
	int Index;

	N->Len = N->Len * Scale;
	N->Scale = N->Scale * Scale;

	for(Index=0;Index<N->NoNodes;Index++)
		ScaleNode(N->NList[Index], Scale);
}

void	RunNode(DATA *Data, int NID, double Scale, int BL)
{
	NODE N;
/*
	printf("Scale\t%d\t%f\t", NID, Scale);
	if(BL == TRUE)
		printf("Branch\n");
	else
		printf("Node\n");
*/
	N = GetNode(Data, NID);
	
	if(N == NULL)
	{
		printf("error getting node %d\n", NID);
		exit(0);
	}

	if(N->Ans == NULL)
	{
		printf("Root is being scaled.\n");
		exit(1);
	}
	
	if(BL == TRUE)
	{
		N->Len = N->Len * Scale;
		N->Scale = N->Scale * Scale;
		return;
	}

	ScaleNode(N, Scale);
}

void	UpDateTaxaToRoot(NODE N)
{
	double	RtoT;
	TAXA	*Taxa;

	Taxa = N->Taxa;

	RtoT = GetNodeToRoot(N);

	Taxa->MeanRT += RtoT;
	Taxa->SDRT += RtoT * RtoT;
}

void	UpDateRootToTip(DATA *Data)
{
	int		Index;
	NODE	N;


	for(Index=0;Index<Data->NoNodes;Index++)
	{
		N = Data->NodeList[Index];
		if(N->Tip == TRUE)
			UpDateTaxaToRoot(N);
	}
}


void	UpDateBLInfo(DATA *Data, int SampleNo)
{
	int		Index;
	BLINFO	*BLInfo;
	NODE	N;

	BLInfo = Data->BLInfo;

	for(Index=0;Index<Data->NoNodes;Index++)
	{
		N = Data->NodeList[Index];
		N->BLList[SampleNo] = N->Len;

		BLInfo->Mean[Index] += N->Len;
		BLInfo->SD[Index] += N->Len * N->Len;
		BLInfo->MeanScale[Index] += N->Scale;

		if(N->Len != BLInfo->BL[Index])
			BLInfo->NoDiff[Index]++;
	}

	UpDateRootToTip(Data);

	BLInfo->No++;
}

void	FinalizeBLInfo(DATA *Data)
{
	int		Index;
	BLINFO	*BLInfo;

	BLInfo = Data->BLInfo;

	for(Index=0;Index<Data->NoNodes;Index++)
	{
		BLInfo->SD[Index] = CalcSD(BLInfo->Mean[Index], BLInfo->SD[Index], BLInfo->No);
		BLInfo->Mean[Index] = BLInfo->Mean[Index] / BLInfo->No;
		BLInfo->MeanScale[Index] = BLInfo->MeanScale[Index] / BLInfo->No;
	}
}

void	FinalizeTaxaRtoT(DATA *Data)
{
	int		Index;
	NODE	N;
	TAXA	*Taxa;

	for(Index=0;Index<Data->NoNodes;Index++)
	{
		N = Data->NodeList[Index];
		if(N->Tip == TRUE)
		{
			Taxa = N->Taxa;
			Taxa->SDRT = CalcSD(Taxa->MeanRT, Taxa->SDRT, Data->BLInfo->No);
			Taxa->MeanRT = Taxa->MeanRT / Data->BLInfo->No;
		}
	}
}

void	Finalize(DATA *Data)
{
	FinalizeBLInfo(Data);
	FinalizeTaxaRtoT(Data);
	FinalizeSPrams(Data);
}

int		NodeSeenBefor(int NodeID, int *List, int Size)
{
	int Index;

	for(Index=0;Index<Size;Index++)
		if(List[Index] == NodeID)
			return TRUE;

	return FALSE;
}

double	GetTotalScaledLength(DATA *Data)
{
	double Ret;
	int Index;
	NODE N;

	Ret = 0;
	for(Index=0;Index<Data->NoNodes;Index++)
	{
		N = Data->NodeList[Index];
		Ret += N->Scale * N->Len;
	//	Ret += N->Len;
	//	Ret += N->Scale;
	}

	return Ret;
}

void	Normalise(double Sigma, DATA *Data)
{
	int Index;
	NODE N;

	printf("====================================================================================================================\n");

//	Sigma = 1.1;

	printf("Sigma:\t%f\n", Sigma);
//	Sigma = log(Sigma);
	printf("Log Sigma:\t%f\n", Sigma);



	for(Index=0;Index<Data->NoNodes;Index++)
	{
		N = Data->NodeList[Index];


		printf("%d\t", Index);
		printf("%f\t", N->TLen);
		printf("Scale:\t%f\t", N->Scale);
		printf("%f\t", N->Len);
		printf("\n");
	}

//	exit(0);
}

void	RunLine(DATA *Data, int SampleNo, int Tokes, char **Passed)
{
	int Index, No, OS, ID, BLen;
	double Scale, Alpha, Sigma;
	int	*SNodes;
	int NoSeen;

	ReSetTree(Data);

	Data->TreeNo = atoi(Passed[0]);

	No = atoi(Passed[3]);
	Alpha = atof(Passed[4]);
	Sigma = atof(Passed[5]);

	Data->Sig2[SampleNo] = Sigma;

	SNodes = (int*)malloc(sizeof(int) * No);
	if(SNodes == NULL)
		MallocErr();

	NoSeen = 0;

	OS = 7;

	for(Index=0;Index<No;Index++)
	{
		ID = atoi(Passed[OS]);

		Scale = atof(Passed[OS+1]);
		if(strcmp(Passed[OS+3], "Node") == 0)
			BLen = FALSE;
		else
			BLen = TRUE;

	//	printf("Thing:\t%s\t%f\n", Passed[OS+3], Scale);

		RunNode(Data, ID, Scale, BLen);
/*
		if(NodeSeenBefor(ID, SNodes, NoSeen) == TRUE)
		{
			printf("On run %d Node %d has been scaled more than once.\n", Data->TreeNo, ID);
			exit(0);
		}
		SNodes[NoSeen] = ID;
		NoSeen++;
*/
		OS += 4;
	}

//	exit(0);

//	printf("Scale\t%f\t%f\n", Sigma, GetTotalScaledLength(Data));

//	if(Data->Opt->Normalise == TRUE)
//		Normalise(Sigma, Data);

	free(SNodes);

	UpDateBLInfo(Data, SampleNo);
}

void	WTree(NODE N, double Cat, double DistSF, int *No, double *List)
{
	int Index;

	if((Cat > DistSF) && (Cat < DistSF + N->TLen))
	{
		List[*No] = N->Scale;
		(*No)++;
		return;
	}

	for(Index=0;Index<N->NoNodes;Index++)
		WTree(N->NList[Index], Cat, DistSF+N->TLen, No, List);
}


double	SMean(int No, double *List)
{
	double Sum;
	int Index;

	Sum = 0;
	for(Index=0;Index<No;Index++)
		Sum += List[Index];

	return Sum / No;
}

void	WandTree(DATA *Data)
{
	int		CIndex;
	int		RIndex;
	RATES	*Rates;
	int		No;
	double	Mean;

	for(RIndex=0;RIndex<Data->Opt->NoRates;RIndex++)
	{
		Rates = Data->Opt->Rates[RIndex];

		for(CIndex=0;CIndex<Rates->NoCats;CIndex++)
		{
			No = 0;

			if(Rates->Cats[CIndex] >= Rates->StatPos)
			{
				WTree(Rates->StartNode, Rates->Cats[CIndex] - 0.000000000001, Rates->StatPos, &No, Rates->Rate);
				Mean = SMean(No, Rates->Rate);
			}

			if(No == 0)
			{
				Mean = 1;
				No = 0;
			}

			Rates->MeanRates[CIndex] += Mean;
			Rates->SDRates[CIndex] += Mean * Mean;
			Rates->NoBL[CIndex] = No;

	//		for(Index=0;Index<No;Index++)
	//			printf("Tree %d:\t%f\t%f\n", Data->TreeNo, Rates->Cats[CIndex], Rates->Rate[Index]);
	//		printf("Tree %d - Mean:\t%f\t%f\n", Data->TreeNo, Rates->Cats[CIndex], SMean(No, Rates->Rate));
		}
	}
}


int		GetNodePos(DATA *Data, int ID)
{
	int Index;

	for(Index=0;Index<Data->NoNodes;Index++)
	{
		if(Data->NodeList[Index]->ID == ID)
			return Index;
	}

	printf("Could not find node ID %d in list.\n", ID);
	exit(0);
	return -1;
}




void	AddScalePram(DATA *Data, int ID, int BLen, double Scale)
{
	int		Pos;
	SPRAM	*SPram;

	Pos = GetNodePos(Data, ID);

	SPram = Data->NodeList[Pos]->SPram;

	if(BLen == TRUE)
	{
		SPram->MeanBL += Scale;
		SPram->SDBL += Scale * Scale;
		SPram->NoBL++;
	}
	else
	{
		SPram->MeanNode += Scale;
		SPram->SDNode += Scale * Scale;
		SPram->NoNode++;
	}
}

void	RunScalePram(DATA *Data, int Tokes, char **Passed)
{
	int Index, OS, ID, BLen, No;
	double Scale;

	No = atoi(Passed[3]);
	OS = 7;

	for(Index=0;Index<No;Index++)
	{
		ID = atoi(Passed[OS]);

		Scale = atof(Passed[OS+1]);
		if(strcmp(Passed[OS+3], "Node") == 0)
			BLen = FALSE;
		else
			BLen = TRUE;

		AddScalePram(Data, ID, BLen, Scale);

		OS += 4;
	}
}

void	FinalSPrams(SPRAM* SP)
{
	SP->SDNode = CalcSD(SP->MeanNode, SP->SDNode, SP->NoNode);
	if(SP->NoNode != 0)
		SP->MeanNode = SP->MeanNode / SP->NoNode;
	else
		SP->MeanNode = 1;

	SP->SDBL = CalcSD(SP->MeanBL, SP->SDBL, SP->NoBL);
	if(SP->NoBL != 0)
		SP->MeanBL = SP->MeanBL / SP->NoBL;
	else
		SP->MeanBL = 1;
}

void	FinalizeSPrams(DATA *Data)
{
	SPRAM*	ScaleP;
	int		Index;

	for(Index=0;Index<Data->NoNodes;Index++)
	{
		ScaleP = Data->NodeList[Index]->SPram;
		FinalSPrams(ScaleP);
	}
}

void	BuildScalePrams(DATA *Data)
{
	char *Buffer, **Passed;
	int Tokes;


	Buffer = (char*)malloc(sizeof(char) * BUFFERSIZE);
	Passed = (char**)malloc(sizeof(char*) * BUFFERSIZE);
	if((Buffer == NULL) || (Passed == NULL))
		MallocErr();

	while(fgets(Buffer, BUFFERSIZE, Data->DFile) != NULL)
	{
		Tokes = MakeArgv(Buffer, Passed, BUFFERSIZE);
		if(Tokes > 1)
		{
			if(atoi(Passed[0]) >= Data->Opt->StartIt)
				RunScalePram(Data, Tokes, Passed);
		}
	}

	free(Buffer);
	free(Passed);
}

void	SetFilePoint(DATA *Data)
{
	char *Buffer;
	int		Index, No;

	Buffer = (char*)malloc(sizeof(char) * BUFFERSIZE);
	if(Buffer == NULL)
		MallocErr();

	if(Data->DFile != NULL)
		fclose(Data->DFile);

	Data->DFile = OpenRead(Data->FName);

	fgets(Buffer, BUFFERSIZE, Data->DFile);

	No = atoi(Buffer);

	for(Index=0;Index<No;Index++)
		fgets(Buffer, BUFFERSIZE, Data->DFile);

	fgets(Buffer, BUFFERSIZE, Data->DFile);

	No = atoi(Buffer);
	for(Index=0;Index<No;Index++)
		fgets(Buffer, BUFFERSIZE, Data->DFile);

	fgets(Buffer, BUFFERSIZE, Data->DFile);

	free(Buffer);
}

void	RunFile(DATA *Data)
{
	char *Buffer, **Passed;
	int		Tokes, SampleNo;

	Buffer = (char*)malloc(sizeof(char) * BUFFERSIZE);
	Passed = (char**)malloc(sizeof(char*) * BUFFERSIZE);
	if((Buffer == NULL) || (Passed == NULL))
		MallocErr();

//	fgets(Buffer, BUFFERSIZE, Data->DFile);
	SetFilePoint(Data);
	BuildScalePrams(Data);
	
	SetFilePoint(Data);
	SampleNo = 0;
	while(fgets(Buffer, BUFFERSIZE, Data->DFile) != NULL)
	{
		Tokes = MakeArgv(Buffer, Passed, BUFFERSIZE);
		if(Tokes > 1)
		{
			if(atoi(Passed[0]) >= Data->Opt->StartIt)
			{
				RunLine(Data, SampleNo, Tokes, Passed);
				WandTree(Data);
				SampleNo++;
			//	Data->NoTrees++;
			}
		}
	}

	ReSetTree(Data);

	fclose(Data->DFile);
	Finalize(Data);

	free(Buffer);
	free(Passed);
}

void	PrintTaxaToRoot(DATA *Data)
{
	int		Index;
	NODE	N;
	TAXA	*Taxa;

	ReSetTree(Data);

	printf("Taxa\t%d\n", Data->NoTaxa);

	printf("Taxa Name\tNodes To Root\tTip To Root Len\tMean RT\tSD RT\tSum Scale RT\n");

	for(Index=0;Index<Data->NoNodes;Index++)
	{
		N = Data->NodeList[Index];
		if(N->Tip == TRUE)
		{
			Taxa = N->Taxa;

			printf("%s\t", Taxa->Name);
			printf("%d\t", GetNoNodes(N));
			printf("%f\t", GetNodeToRoot(N));
			printf("%f\t", Taxa->MeanRT);
			printf("%f\t", Taxa->SDRT);
			printf("%f\t", SumScaleToRoot(Data, N));

			printf("\n");
		}
	}

	printf("\n\n");
}


void	PrintScalePara(DATA *Data)
{
	int	Index;
	SPRAM* SP;

	printf("Scaling Parameters\n");
	printf("Node ID\tPct Node\tPct Branch\tNode Mean\tNode SD\tBranch Mean\tBranch SD\n");

	for(Index=0;Index<Data->NoNodes;Index++)
	{
		SP = Data->NodeList[Index]->SPram;

		printf("%d\t", SP->Node->ID);

		if(SP->NoNode != 0)
			printf("%f\t", (double)SP->NoNode / Data->NoTrees);
		else
			printf("0.0\t");


		if(SP->NoBL != 0)
			printf("%f\t", (double)SP->NoBL / Data->NoTrees);
		else
			printf("0.0\t");

		printf("%f\t%f\t", SP->MeanNode, SP->SDNode);
		printf("%f\t%f\t", SP->MeanBL, SP->SDBL);



		printf("\n");
	}
}

void	PrintCumeScaleParam(DATA *Data)
{
	int		Index;
	int		NIndex;
	NODE	N;
	NODE	*NList;
	int		Size;
	SPRAM*	SP;
	double	PctScale;

	NList = (NODE*)malloc(sizeof(NODE) * Data->NoNodes);
	if(NList == NULL)
		MallocErr();

	printf("Prints how each node is scale and by which other nodes.\n");

	printf("Node ID:	The unique identifier for each node.\n");
	printf("Pct BL Scale:	The percentage of time the branch length is scale\n");
	printf("Mean BL Scale:	The mean scalar for the branch length. \n");
	printf("SD BL Scale:	The standard deviation of the branch length scalar \n");

	printf("No Scales:	Number of downstream nodes\n");
	printf("DS Node ID:	unique identities of downstream nodes\n");


	printf("Node ID\tPct BL Scale\tMean BL Scale\tSD BL Scale\tNo Scales\tDS Node ID\n");

	for(Index=0;Index<Data->NoNodes;Index++)
	{
		N = Data->NodeList[Index];
		SP = N->SPram;

		printf("%d\t", N->ID);
		printf("%f\t%f\t%f\t", (double)SP->NoBL / Data->NoTrees, SP->MeanBL, SP->SDBL);

		Size = 0;
		while(N != NULL)
		{
			NList[Size++] = N;
			N = N->Ans;
		}

		PctScale = 0;
		for(NIndex=0;NIndex<Size;NIndex++)
		{
			N = NList[NIndex];
			PctScale +=  (double)N->SPram->NoNode / Data->NoTrees;
		}

		printf("%d\t", Size);
		for(NIndex=0;NIndex<Size;NIndex++)
		{
			N = NList[NIndex];
			printf("%d\t", N->ID);
		}

		printf("\n");
	}

	printf("\n\n\n\n");

	free(NList);
}

int cstring_cmp(const void *a, const void *b)
{
    char **ia = (char **)a;
	char **ib = (char **)b;
	return strcmp(*ia, *ib);
}

void	PrintTaxaList(NODE N, DATA *Data)
{
	int Index;
	char	**TList;
	TAXA		*T;

	TList = (char**)malloc(sizeof(char*) * N->NoPart);
	if(TList == NULL)
		MallocErr();

	for(Index=0;Index<N->NoPart;Index++)
	{
		T = GetTaxa(N->Part[Index], Data);
		TList[Index] = T->Name;
	}
	
	qsort(TList, N->NoPart, sizeof(char*), cstring_cmp);

	for(Index=0;Index<N->NoPart-1;Index++)
		printf("%s,", TList[Index]);
	printf("%s", TList[Index]);
	
	free(TList);
}

void	PrintHTMLKernel(NODE N, KERNELD *KD, double Mode)
{
	if(N->Len == 0)
	{
		printf("\t\t<td>0.0</td>\n");
		printf("\t\t<td>0.0</td>\n");
		printf("\t\t<td>0.0</td>\n");
		printf("\t\t<td>0.0</td>\n");
		return;
	}

	printf("\t\t<td>%f</td>\n", KD->Mean);
	printf("\t\t<td>%f</td>\n", Mode);
	printf("\t\t<td>%f</td>\n", KD->Median);
	printf("\t\t<td>%f</td>\n", KD->SD);
}

void	HTMLPrintOutput(DATA *Data)
{
	int		Index;
	NODE	N;
	BLINFO	*BLInfo;
	
	

	// prints how each node is caled and by whom.
//	PrintCumeScaleParam(Data);

//	PrintScalePara(Data);

	BLInfo = Data->BLInfo;

//	printf("<table style=\"width:100%\">\n");
	printf("<table>\n");
	printf("\t<tr>\n");
	printf("\t\t<th>Node ID</th>\n");
	printf("\t\t<th>No Taxa</th>\n");
	printf("\t\t<th>Original BL</th>\n");
	
	printf("\t\t<th>No different from original</th>\n");
	printf("\t\t<th>Pct time scaled</th>\n");

	printf("\t\t<th>Mean Scalar</th>\n");
	printf("\t\t<th>Mode Scalar</th>\n");
	printf("\t\t<th>Median Scalar</th>\n");
	printf("\t\t<th>SD Scalar</th>\n");
		
	printf("\t\t<th>Mean SigV</th>\n");
	printf("\t\t<th>Mode SigV</th>\n");
	printf("\t\t<th>Median SigV</th>\n");
	printf("\t\t<th>SD SigV</th>\n");
		
	printf("\t\t<th>Mean DeltaB</th>\n");
	printf("\t\t<th>Mode DeltaB</th>\n");
	printf("\t\t<th>Median DeltaB</th>\n");
	printf("\t\t<th>SD DeltaB</th>\n");

	printf("\t\t<th>Mean DeltaV</th>\n");
	printf("\t\t<th>Mode DeltaV</th>\n");
	printf("\t\t<th>Median DeltaV</th>\n");
	printf("\t\t<th>SD DeltaV</th>\n");
		
	printf("\t\t<th>Taxa List</th>\n");
	printf("\t</tr>\n");
	
	for(Index=0;Index<Data->NoNodes;Index++)
	{
		printf("\t<tr>\n");

		N = Data->NodeList[Index];


		printf("\t\t<td>%d</td>\n", N->ID);
		printf("\t\t<td>%d</td>\n", N->NoPart);
		printf("\t\t<td>%f</td>\n", N->Len);


		printf("\t\t<td>%d</td>\n", BLInfo->NoDiff[Index]);
		printf("\t\t<td>%f</td>\n", 100*((double)BLInfo->NoDiff[Index] / BLInfo->No));
		
		PrintHTMLKernel(N, N->KDR, N->ModeR);
		PrintHTMLKernel(N, N->KDSigV, N->ModeSigV);
		PrintHTMLKernel(N, N->KDDeltaB, N->ModeDeltaB);
		PrintHTMLKernel(N, N->KDDeltaV, N->ModeDeltaV);
		
		printf("\t\t<td>");
		PrintTaxaList(N, Data);
		printf("</td>\n");

		printf("\t</tr>\n");
	}

	printf("</table>\n");
}

void	PrintKernel(NODE N, KERNELD *KD, double Mode)
{
	if(N->Len == 0)
	{
		printf("0.0\t");
		printf("0.0\t");
		printf("0.0\t");
		printf("0.0\t");
		return;
	}

	printf("%f\t", KD->Mean);
	printf("%f\t", Mode);
	printf("%f\t", KD->Median);
	printf("%f\t", KD->SD);
}

void	PrintExtaInfo(NODE N, int Size, double Len)
{
	int Index;

	for(Index=0;Index<Size;Index++)
		printf("%f\t", N->BLList[Index] / Len);
}

void	PrintOutput(DATA *Data)
{
	int		Index;
	double	Mode;
	NODE	N;
	BLINFO	*BLInfo;

	// prints how each node is caled and by whom.
//	PrintCumeScaleParam(Data);

//	PrintScalePara(Data);

	BLInfo = Data->BLInfo;


	printf("Node ID\t");
	printf("No Taxa\t");
	printf("Original BL\t");
		
	printf("No different from original\t");
	printf("Pct time scaled\t");

	printf("Mean Scalar\t");
	printf("Mode Scalar\t");
	printf("Median Scalar\t");
	printf("SD Scalar\t");

	printf("Mean SigV\t");
	printf("Mode SigV\t");
	printf("Median SigV\t");
	printf("SD SigV\t");

	printf("Mean DeltaB\t");
	printf("Mode DeltaB\t");
	printf("Median DeltaB\t");
	printf("SD DeltaB\t");

	printf("Mean DeltaV\t");
	printf("Mode DeltaV\t");
	printf("Median DeltaV\t");
	printf("SD DeltaV\t");


	printf("Taxa List\t");

	printf("\n");

	for(Index=0;Index<Data->NoNodes;Index++)
	{
		N = Data->NodeList[Index];

		printf("%d\t%d\t%f\t", N->ID, N->NoPart, N->Len);

		printf("%d\t", BLInfo->NoDiff[Index]);
		printf("%f\t", 100*((double)BLInfo->NoDiff[Index] / BLInfo->No));
		

		PrintKernel(N, N->KDR, N->ModeR);
		PrintKernel(N, N->KDSigV, N->ModeSigV);
		PrintKernel(N, N->KDDeltaB, N->ModeDeltaB);
		PrintKernel(N, N->KDDeltaV, N->ModeDeltaV);
				
		PrintTaxaList(N, Data);
		printf("\t");
//		Save the rates. 
//		PrintExtaInfo(N, BLInfo->No, N->Len);

		printf("\n");
	}

	printf("\n\n\n\n");

//	PrintTaxaToRoot(Data);
}


void	WriteNode(FILE *Out, NODE N, DATA *Data)
{
	int		Pos;
	BLINFO	*BLInfo;
	int		Index;

	BLInfo = Data->BLInfo;

	Pos = N->Pos;

	if(N->Tip == TRUE)
	{
		fprintf(Out, "%d:%f", N->Taxa->No, BLInfo->Mean[Pos]);
		return;
	}

	fprintf(Out, "(");

	for(Index=0;Index<N->NoNodes-1;Index++)
	{
		WriteNode(Out, N->NList[Index], Data);
		fprintf(Out, ",");
	}
	WriteNode(Out, N->NList[Index], Data);

	if(N->Ans != NULL)
		fprintf(Out, "):%f", BLInfo->Mean[Pos]);
	else
		fprintf(Out, ")");
}

void	SaveTree(DATA *Data)
{
	FILE	*Out;
	char	*Buffer;
	int		Index;
	TAXA	*Taxa;

	Buffer = (char*)malloc(sizeof(char) * BUFFERSIZE);
	if(Buffer == NULL) MallocErr();

	sprintf(Buffer, "%s.trees", Data->FName);

	Out = OpenWrite(Buffer);

	fprintf(Out, "#Nexus\n\n");
	fprintf(Out, "Begin trees;\n");
	fprintf(Out, "\tTranslate\n");

	for(Index=0;Index<Data->NoTaxa-1;Index++)
	{
		Taxa = Data->TaxaList[Index];
		fprintf(Out, "\t\t%d\t%s,\n", Taxa->No, Taxa->Name);
	}
	Taxa = Data->TaxaList[Index];
	fprintf(Out, "\t\t%d\t%s;\n", Taxa->No, Taxa->Name);

	fprintf(Out, "\t\ttree MeanTree = ");
	WriteNode(Out, Data->Root, Data);

	fprintf(Out, ";\nend;\n");

	free(Buffer);
	fclose(Out);
}

void	PrintSlices(DATA *Data)
{
	OPTIONS *Opt;
	int		Index;
	SLICE	*Slice;

	Opt = Data->Opt;

	printf("Non phylogenetic slice : do not use.\n");

	printf("Slices\t%d\n", Opt->NoSlices);

	for(Index=0;Index<Opt->NoSlices;Index++)
	{
		Slice = Opt->Slices[Index];

		printf("Slice\t%s\t%f\n", Slice->Name, Slice->Point);
	}
}

void	PrintRate(DATA *Data, RATES *Rates)
{
	int		Index;
	double	SD;

	printf("Rate Name\t%s\n", Rates->Name);
	printf("\tNo Taxa\t%d\n", Rates->StartNode->NoPart);
	printf("\tNo Cats\t%d\n", Rates->NoCats);
	printf("\tStart Deepth\t%f\n", Rates->StatPos);

	printf("\t%s No BL\t", Rates->Name);
	for(Index=0;Index<Rates->NoCats;Index++)
		printf("%d\t", Rates->NoBL[Index]);
	printf("\n");

	printf("\t%s Means\t", Rates->Name);
	for(Index=0;Index<Rates->NoCats;Index++)
		printf("%f\t", (double)Rates->MeanRates[Index] / Data->NoTrees);
	printf("\n");

	printf("\t%s SD\t", Rates->Name);
	for(Index=0;Index<Rates->NoCats;Index++)
	{
		SD = CalcSD(Rates->MeanRates[Index], Rates->SDRates[Index], Data->NoTrees);
		printf("%f\t", SD);
	}
	printf("\n");
}

void	PrintRates(DATA *Data)
{
	int		Index;
	RATES	*Rates;


	printf("Printing Rates:\t%d\n", Data->Opt->NoRates);

	Rates = Data->Opt->Rates[0];
	printf("\t%s Cats\t", Rates->Name);
	for(Index=0;Index<Rates->NoCats;Index++)
		printf("%f\t", Rates->Cats[Index]);
	printf("\n");

	for(Index=0;Index<Data->Opt->NoRates;Index++)
		PrintRate(Data, Data->Opt->Rates[Index]);
}

void	SetConBLScale(DATA *Data)
{
	int Index, Pos;
	BLINFO		*BLInfo;
	NODE		N;

	ReSetTree(Data);
	BLInfo = Data->BLInfo;

	for(Index=0;Index<Data->NoNodes;Index++)
	{
		N = Data->NodeList[Index];
		Pos = N->Pos;
		N->Scale = BLInfo->MeanScale[Pos];
	}
}

int		NoTipsBelow(NODE N)
{
	int Index;
	int No;

	if(N->Tip == TRUE)
		return 1;

	No = 0;
	for(Index=0;Index<N->NoNodes;Index++)
		No += NoTipsBelow(N->NList[Index]);

	return No;
}

void	GetSumBLScale(DATA *Data, NODE N, int *NoNodes, double *SumScale, double *SSE)
{
	int Index;

	(*NoNodes)++;
	(*SumScale) += Data->BLInfo->MeanScale[N->Pos];
	(*SSE) += Data->BLInfo->MeanScale[N->Pos] * Data->BLInfo->MeanScale[N->Pos];

	for(Index=0;Index<N->NoNodes;Index++)
		GetSumBLScale(Data, N->NList[Index], NoNodes, SumScale, SSE);
}

void	PrintNodeScale(DATA *Data, NODE N, int *NoNodes, double *SumScale)
{
	int Index;

	(*NoNodes)++;
	(*SumScale) += Data->BLInfo->MeanScale[N->Pos];


	printf("Got\t%d\t%f\t%d\t%f\n", *NoNodes, *SumScale, N->ID, Data->BLInfo->MeanScale[N->Pos]);

	for(Index=0;Index<N->NoNodes;Index++)
		PrintNodeScale(Data, N->NList[Index], NoNodes, SumScale);
}

void	MakeTaxaList(NODE N, NODE *TList, int *Pos)
{
	int Index;

	if(N->Tip == TRUE)
	{
		TList[*Pos] = N;
		(*Pos)++;
		return;
	}

	for(Index=0;Index<N->NoNodes;Index++)
		MakeTaxaList(N->NList[Index], TList, Pos);
}


void	WriteBigRatioNode(FILE *TFile, NODE ONode, NODE N, DATA* Data, int Scale)
{
	int		Pos;
	BLINFO	*BLInfo;
	int		Index;

	BLInfo = Data->BLInfo;

	Pos = N->Pos;

	if(N->Tip == TRUE)
	{
		if(Scale == TRUE)
			fprintf(TFile, "%d:%f", N->Taxa->No, N->Len * N->Scale);
		else
			fprintf(TFile, "%d:%f", N->Taxa->No, N->Len);

		return;
	}

	fprintf(TFile, "(");

	for(Index=0;Index<N->NoNodes-1;Index++)
	{
		WriteBigRatioNode(TFile, ONode, N->NList[Index], Data, Scale);
		fprintf(TFile, ",");
	}
	WriteBigRatioNode(TFile, ONode, N->NList[Index], Data, Scale);

	if(N != ONode )
	{
		if(Pos != -1)
		{
			if(Scale == TRUE)
				fprintf(TFile, "):%f", N->Len * N->Scale);
			else
				fprintf(TFile, "):%f", N->Len);
		}
		else
			fprintf(TFile, "):0.000");
	}
	else
		fprintf(TFile, ")");
}


void	PrintBigRatioNode(DATA *Data, NODE N, int NoTaxa, char *Dir)
{
	char*	Buffer;
	FILE	*TFile;
	NODE	*TList;
	int		Pos, Index;

	Buffer = (char*)malloc(sizeof(char) * BUFFERSIZE);
	if(Buffer == NULL)
		MallocErr();

	sprintf(Buffer, "%s\\ScaleDiff-%06d.trees", Dir, N->ID);
	TFile = OpenWrite(Buffer);
	free(Buffer);

	fprintf(TFile, "#Nexus\n\n");
	fprintf(TFile, "Begin trees;\n");
	fprintf(TFile, "\tTranslate\n");

	TList = (NODE*)malloc(sizeof(NODE) * NoTaxa);
	if(TList == NULL)
		MallocErr();

	Pos = 0;
	MakeTaxaList(N, TList, &Pos);

	for(Index=0;Index<NoTaxa-1;Index++)
		fprintf(TFile, "\t\t%d\t%s,\n", TList[Index]->Taxa->No, TList[Index]->Taxa->Name);
	fprintf(TFile, "\t\t%d\t%s;\n", TList[Index]->Taxa->No, TList[Index]->Taxa->Name);

	fprintf(TFile, "\t\ttree Node-%06d = ", N->ID);
	WriteBigRatioNode(TFile, N, N, Data, TRUE);
	fprintf(TFile, ";\n");
	fprintf(TFile, "\t\ttree Node-%06d = ", N->ID);
	WriteBigRatioNode(TFile, N, N, Data, FALSE);

	fprintf(TFile, ";\nend;\n");


	free(TList);
	fclose(TFile);
}

void	PrintScaleNodeList(DATA *Data, NODE N)
{
	int Index;

	printf("%f\t", Data->BLInfo->MeanScale[N->Pos]);

	for(Index=0;Index<N->NoNodes;Index++)
		PrintScaleNodeList(Data, N->NList[Index]);
}

void	AnalyseRatioNode(DATA *Data, NODE N, char *Dir)
{
	int		NoBL;
	double	AveScale, Scale, Ratio, SSE, SD;
	int		Index, NoTaxa;

	NoTaxa = NoTipsBelow(N);

	Scale = Data->BLInfo->MeanScale[N->Pos];

	NoBL = 0;
	AveScale = 0;
	SSE = 0;

	for(Index=0;Index<N->NoNodes;Index++)
		GetSumBLScale(Data, N->NList[Index], &NoBL, &AveScale, &SSE);

	SD = CalcSD(AveScale, SSE, NoBL);
	AveScale = AveScale / NoBL;
	Ratio = Scale / AveScale;

	printf("BLS Scale\t%d\t", N->ID);
	printf("%f\t%f\t", Ratio, Scale);
	printf("%f\t%d\t%d\t", AveScale, NoBL, NoTaxa);
	printf("%f\t", SD);

	printf("|\t");
	for(Index=0;Index<N->NoNodes;Index++)
		PrintScaleNodeList(Data, N->NList[Index]);
	printf("|\t");

	printf("\n");

	PrintBigRatioNode(Data, N, NoTaxa, Dir);
}

int		CheckBigRatio(DATA *Data, NODE N, char *Dir)
{
	int		NoBL;
	double	AveScale, Scale, Ratio, SSE, SD;
	int		Index, NoTaxa;

	if(N == Data->Root)
		return FALSE;

	if(N->Tip == TRUE)
		return FALSE;

	NoTaxa = NoTipsBelow(N);
	if(NoTaxa < Data->Opt->MinRatioSize)
		return FALSE;

	Scale = Data->BLInfo->MeanScale[N->Pos];

	NoBL = 0;
	AveScale = 0;
	SSE = 0;

	for(Index=0;Index<N->NoNodes;Index++)
		GetSumBLScale(Data, N->NList[Index], &NoBL, &AveScale, &SSE);

	SD = CalcSD(AveScale, SSE, NoBL);
	AveScale = AveScale / NoBL;
	Ratio = Scale / AveScale;

	if(!((Ratio >= Data->Opt->Ratio) || (Ratio <= 1.0 / Data->Opt->Ratio)))
		return FALSE;

	AnalyseRatioNode(Data, N, Dir);

	return TRUE;
}

void	FindBigRatio(DATA *Data, char *Dir)
{
	int NIndex;
	NODE N;

	for(NIndex=0;NIndex<Data->NoNodes;NIndex++)
	{
		N = Data->NodeList[NIndex];
		CheckBigRatio(Data, N, Dir);
	}
}

void	PrintUserRatioNodes(DATA *Data, char *Dir)
{
	int			Index;
	OPTIONS		*Opt;
	NODE		N;
	NODERATIO	*NodeRatio;

	Opt = Data->Opt;

	for(Index=0;Index<Opt->NoNodeRatio;Index++)
	{
		NodeRatio = Opt->NodeRatio[Index];

		N = GetNodeFromTaxaList(Data, NodeRatio->NoTaxa, NodeRatio->TNames);
		if(N == NULL)
			printf("No valid nodes for %s\n", NodeRatio->Name);
		else
		{
			printf("Use Def Node\t%s\t", NodeRatio->Name);
			AnalyseRatioNode(Data, N, Dir);
		}
	}
}

double	GetMaxRate(DATA *Data)
{
	BLINFO *BLInfo;
	int Index;
	double Ret;

	BLInfo = Data->BLInfo;

	Ret = BLInfo->MeanScale[0];

	for(Index=1;Index<Data->NoNodes;Index++)
		if(BLInfo->MeanScale[Index] > Ret)
			Ret = BLInfo->MeanScale[Index];

	return Ret;
}

double	GetMinRate(DATA *Data)
{
	BLINFO *BLInfo;
	int Index;
	double Ret;

	BLInfo = Data->BLInfo;

	Ret = BLInfo->MeanScale[1];

	for(Index=1;Index<Data->NoNodes;Index++)
		if(BLInfo->MeanScale[Index] < Ret)
			Ret = BLInfo->MeanScale[Index];

	return Ret;
}

void	GetColourB(double Min, double Max, double Rate, int *R, int *G, int *B)
{
	double dR, dG, dB, Range, ORate;
	
	if(Rate < Min)
	{
		*R = 0;
		*G = 0;
		*B = 0;
		return;
	}

	ORate = Rate;
	
	if(1.0 / Max < Min)
	{
		Range = log(1.0 / Max);
	}
	else
		Range = log(Min);

	if(Rate > 1)
		Rate = log(1.0 / Rate);
	else
		Rate = log(Rate);


	dB = 0;
	dG = 0;
	dR = 0;

	if(ORate >= 1)
		dR = (Rate / Range) * 255;
	else
		dB = (Rate / Range) * 255;

	*R = (int)floor(dR + 0.5);
	*G = (int)floor(dG + 0.5);
	*B = (int)floor(dB + 0.5);
}

void	OutColourFileB(DATA *Data, NODE N, FILE *Out, double Rate, double Min, double Max)
{
	int R,G,B;
	int Index;
	TAXA *T;

	GetColourB(Min, Max, Rate, &R, &G, &B);
		
	fprintf(Out, "#%02X%02X%02X\t", R, G, B);
	fprintf(Out, "A\t2\t");

	for(Index=0;Index<N->NoPart;Index++)
	{
		T = GetTaxa(N->Part[Index], Data);
		fprintf(Out, "%s ", T->Name);
	}
	fprintf(Out, "\n");
}

void	GetColourA(double Min, double Max, double Rate, int *R, int *G, int *B)
{
	double dR, dG, dB;

	if(Rate < Min)
	{
		*R = 0;
		*G = 0;
		*B = 0;
		return;
	}
	
	Min = log(Min);
	Max = log(Max);
	Rate = log(Rate);
	
	dG = 0;
	dR = (Rate - Min) / ((Max - Min) / 255.0);
	dB = 255.0 - dR;
	
	*R = (int)floor(dR + 0.5);
	*G = (int)floor(dG + 0.5);
	*B = (int)floor(dB + 0.5);
}

void	OutColourFileA(DATA *Data, NODE N, FILE *Out, double Rate, double Min, double Max)
{
	
	int R,G,B;
	int Index;
	TAXA *T;

	GetColourA(Min, Max, Rate, &R, &G, &B);
	
	fprintf(Out, "#%02X%02X%02X\t", R, G, B);
	fprintf(Out, "A\t2\t");

	for(Index=0;Index<N->NoPart;Index++)
	{
		T = GetTaxa(N->Part[Index], Data);
		fprintf(Out, "%s ", T->Name);
	}
	fprintf(Out, "\n");
}

FILE*	MakeFileName(char *Base, char *Ext)
{
	FILE*	Ret;
	char	*Buffer;

	Buffer = (char*)malloc(sizeof(char) * (sizeof(Base) + sizeof(Ext) + 64));
	if(Buffer == NULL)
		MallocErr();
	sprintf(Buffer, "%s%s", Base, Ext);

	Ret = OpenWrite(Buffer);

	free(Buffer);

	return Ret;
}

void	MakeColourTable(DATA *Data, double Min, double Max)
{
	FILE	*Out;
	int		R, G, B;
	double	Rate, Step;

	Out = OpenWrite("ColourList.html");

	fprintf(Out, "<!DOCTYPE html><html><body>\n");

	fprintf(Out, "<div style=\"background-color: #FFFFFF; margin-left: 50px; margin-right: 500px; padding-bottom: 8px; padding-left: 8px; padding-right: 8px; padding-top: 8px;\"> <center>%s.Colour.A.txt</center> </div>\n", Data->FName);

	Step = (1.0 - Min) / 10.0;
	for(Rate=Min;Rate<1;Rate+=Step)
	{
		GetColourA(Min, Max, Rate, &R, &G, &B);
		fprintf(Out, "<div style=\"color: #ffffff;background-color: #%02X%02X%02X; margin-left: 50px; margin-right: 500px; padding-bottom: 8px; padding-left: 8px; padding-right: 8px; padding-top: 8px;\"> <center>%f</center> </div>\n", R, G, B, Rate);
	}

	Rate = 1.0;
	GetColourA(Min, Max, Rate, &R, &G, &B);
	fprintf(Out, "<div style=\"color: #ffffff;background-color: #%02X%02X%02X; margin-left: 50px; margin-right: 500px; padding-bottom: 8px; padding-left: 8px; padding-right: 8px; padding-top: 8px;\"> <center>%f</center> </div>\n", R, G, B, Rate);
	
	Step = (Max - 1.0) / 10.0;
	for(Rate=1+Step;Rate<=Max;Rate+=Step)
	{
		GetColourA(Min, Max, Rate, &R, &G, &B);
		fprintf(Out, "<div style=\"color: #ffffff;background-color: #%02X%02X%02X; margin-left: 50px; margin-right: 500px; padding-bottom: 8px; padding-left: 8px; padding-right: 8px; padding-top: 8px;\"> <center>%f</center> </div>\n", R, G, B, Rate);
	}

	fprintf(Out, "<BR><BR><BR>\n");
	fprintf(Out, "<div style=\"background-color: #FFFFFF; margin-left: 50px; margin-right: 500px; padding-bottom: 8px; padding-left: 8px; padding-right: 8px; padding-top: 8px;\"> <center>%s.Colour.B.txt</center> </div>\n", Data->FName);

	Step = (1.0 - Min) / 10.0;
	for(Rate=Min;Rate<1;Rate+=Step)
	{
		GetColourB(Min, Max, Rate, &R, &G, &B);
		fprintf(Out, "<div style=\"color: #ffffff;background-color: #%02X%02X%02X; margin-left: 50px; margin-right: 500px; padding-bottom: 8px; padding-left: 8px; padding-right: 8px; padding-top: 8px;\"> <center>%f</center> </div>\n", R, G, B, Rate);
	}

	Rate = 1.0;
	GetColourB(Min, Max, Rate, &R, &G, &B);
	fprintf(Out, "<div style=\"color: #ffffff;background-color: #%02X%02X%02X; margin-left: 50px; margin-right: 500px; padding-bottom: 8px; padding-left: 8px; padding-right: 8px; padding-top: 8px;\"> <center>%f</center> </div>\n", R, G, B, Rate);

	Step = (Max - 1.0) / 10.0;
	for(Rate=1+Step;Rate<=Max;Rate+=Step)
	{
		GetColourB(Min, Max, Rate, &R, &G, &B);
		fprintf(Out, "<div style=\"color: #ffffff;background-color: #%02X%02X%02X; margin-left: 50px; margin-right: 500px; padding-bottom: 8px; padding-left: 8px; padding-right: 8px; padding-top: 8px;\"> <center>%f</center> </div>\n", R, G, B, Rate);
	}

/*
	fprintf(Out, "<BR><BR><BR>\n");
	fprintf(Out, "<div style=\"background-color: #FFFFFF; margin-left: 50px; margin-right: 500px; padding-bottom: 8px; padding-left: 8px; padding-right: 8px; padding-top: 8px;\"> <center>%s.Colour.B.txt</center> </div>\n", Data->FName);
	for(Rate=Min;Rate<=Max;Rate+=Step)
	{
		GetColourB(Min, Max, Rate, &R, &G, &B);
		fprintf(Out, "<div style=\"background-color: #%02X%02X%02X; margin-left: 50px; margin-right: 500px; padding-bottom: 8px; padding-left: 8px; padding-right: 8px; padding-top: 8px;\"> <center>%f</center> </div>\n", R, G, B, Rate);
	}
	*/
	fprintf(Out, "</body></html>\n");
	fclose(Out);
}

void	MakeColourFile(DATA *Data)
{
	int		Index;
	double	RateMin, RateMax;
	NODE	N;
	BLINFO	*BLInfo;
	FILE	*OutA, *OutB;
	
	OutA = MakeFileName(Data->FName, ".Colour.A.txt");
	OutB = MakeFileName(Data->FName, ".Colour.B.txt");

	RateMax = GetMaxRate(Data);
	RateMin = GetMinRate(Data);
	
	BLInfo = Data->BLInfo;

	for(Index=0;Index<Data->NoNodes;Index++)
	{
		N = Data->NodeList[Index];
		OutColourFileA(Data, N, OutA, BLInfo->MeanScale[Index], RateMin, RateMax);
		OutColourFileB(Data, N, OutB, BLInfo->MeanScale[Index], RateMin, RateMax);
	}

	fclose(OutA);
	fclose(OutB);

	MakeColourTable(Data, RateMin, RateMax);
//	exit(0);
}

void	AllocSampleSize(DATA *Data)
{
	NODE N;
	int NIndex;

	for(NIndex=0;NIndex<Data->NoNodes;NIndex++)
	{
		N = Data->NodeList[NIndex];
		N->BLList = (double*)malloc(sizeof(double) * Data->NoTrees);
		if(N->BLList == NULL)
			MallocErr();
	}

	Data->Sig2 = (double*)malloc(sizeof(double) * Data->NoTrees);
	if(Data->Sig2 == NULL)
		MallocErr();
}

void	FindNoSamples(DATA *Data)
{
	int Index, Tokes;
	char *Buffer;
	char **Passed;

	Buffer = (char*)malloc(sizeof(char) * BUFFERSIZE);
	Passed = (char**)malloc(sizeof(char*) * BUFFERSIZE);
	if(Buffer == NULL || Passed == NULL)
		MallocErr();

	SetFilePoint(Data);

	Data->NoTrees = 0;
	while(fgets(Buffer, BUFFERSIZE, Data->DFile) != NULL)
	{
		Tokes = MakeArgv(Buffer, Passed, BUFFERSIZE);
		if(Tokes > 1)
		{
			if(atoi(Passed[0]) >= Data->Opt->StartIt)
				Data->NoTrees++;
		}
	}

	fclose(Data->DFile);
	Data->DFile = NULL;

	free(Buffer);
	free(Passed);

	AllocSampleSize(Data);
}

void	SetNodeKernels(NODE N, DATA *Data)
{
	int Index;
	double *R, *SigV, *DeltaB, *DeltaV;

	R = (double*)malloc(sizeof(double) * Data->NoTrees);
	SigV = (double*)malloc(sizeof(double) * Data->NoTrees);
	DeltaB = (double*)malloc(sizeof(double) * Data->NoTrees);
	DeltaV = (double*)malloc(sizeof(double) * Data->NoTrees);

	if(R == NULL || SigV == NULL || DeltaB == NULL || DeltaV == NULL)
		MallocErr();

	for(Index=0;Index<Data->NoTrees;Index++)
	{
		R[Index] = N->BLList[Index] / N->Len;
		SigV[Index] = Data->Sig2[Index] * R[Index];
		DeltaB[Index] = Data->Sig2[Index] * N->Len;
		DeltaV[Index] = SigV[Index] * N->Len;
	}

	N->KDR		= CreateKD(Data->NoTrees, R);
	N->KDSigV	= CreateKD(Data->NoTrees, SigV);
	N->KDDeltaB = CreateKD(Data->NoTrees, DeltaB);
	N->KDDeltaV = CreateKD(Data->NoTrees, DeltaV);

	SetBandWidthSilverman(N->KDR);
	SetBandWidthSilverman(N->KDSigV);
	SetBandWidthSilverman(N->KDDeltaB);
	SetBandWidthSilverman(N->KDDeltaV);

	N->ModeR		= FindMode(N->KDR, 1000);
	N->ModeSigV		= FindMode(N->KDSigV, 1000);
	N->ModeDeltaB	= FindMode(N->KDDeltaB, 1000);
	N->ModeDeltaV	= FindMode(N->KDDeltaV, 1000);

	free(R);
	free(SigV);
	free(DeltaB);
	free(DeltaV);
}

void	SetKernels(DATA *Data)
{
	int Index;
	NODE N;

	#pragma omp parallel for private(N)
	for(Index=0;Index<Data->NoNodes;Index++)
	{
		N = Data->NodeList[Index];
		SetNodeKernels(N, Data);
 	}
}

//./Seq/H/dat_8.1_trans.txt.log.txt.VarRates.txt ./Seq/Options.txt > sout.txt

int main(int argc, char **argv)
{
	DATA *Data;
	
	if((argc != 2) && (argc != 3))
	{
		printf("%s takes an options file and a log file. \n", __FILE__);
		return 0;
	}

	Data = LoadData(argv[1]);
	BuildTree(Data);

	Data->Opt = CreatOptions();

	if(argc == 3)
		LoadOptions(Data, argv[2]);
		
	FindNoSamples(Data);

	RunFile(Data);

	SetKernels(Data);

//	printf("No Trees\t%d\n", Data->NoTrees);
//	printf("No Taxa\t%d\n", Data->NoTaxa);

//	PrintBLInfo(Data);
#ifdef HTML_OUTPUT
	HTMLPrintOutput(Data);
#else
	PrintOutput(Data);
#endif
	exit(0);
//	PrintSlices(Data);

//	PrintRates(Data);

	SaveTree(Data);

//	Taken out of the workshop 3/11/14
	SetConBLScale(Data);
	PhyCorrectCon(Data, OUTPUTDIR);

//	if(Data->Opt->FindRatio == TRUE)
//		FindBigRatio(Data, OUTPUTDIR);

//	PrintUserRatioNodes(Data, OUTPUTDIR);
	
//  this is not safe to use used with multiple istances running
//	when run on the web server. 
//	MakeColourFile(Data);

	FreeData(Data);

	return 0;
}