#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "genlib.h"
#include "PhyCorrect.h"
#include "PPPostProcess.h"

#define PHYCORRUM

extern void SetConBLScale(DATA *Data);

void	RecFreeTree(NODE N)
{
	int Index;

	for(Index=0;Index<N->NoNodes;Index++)
		RecFreeTree(N->NList[Index]);

	if((N->Tip == FALSE) && (N->NList != NULL))
		free(N->NList);

	free(N);
}


NODE	CloneNode(NODE Node)
{
	NODE Ret;

	Ret = (NODE)malloc(sizeof(struct INODE));
	if(Ret == NULL)
		MallocErr();

	Ret->TLen	= Node->TLen;
	Ret->Len	= Node->Len;
	Ret->Scale	= Node->Scale;
	Ret->ID		= Node->ID;
	Ret->Tip	= FALSE;
	Ret->Pos	= Node->Pos;
	Ret->NoPart	= Node->NoPart;
	Ret->Part	= Node->Part;

	Ret->Taxa	= NULL;

	Ret->NList	= NULL;
	Ret->Ans	= NULL;
	Ret->NoNodes= 0;

	Ret->SPram	= NULL;

	return Ret;
}

STREE*	MakeSTree(char *Name)
{
	STREE*	Ret;

	Ret = (STREE*)malloc(sizeof(STREE));
	if(Ret == NULL)
		MallocErr();

	Ret->NoNodes	= 0;
	Ret->NodeList	= NULL;
	Ret->Root		= NULL;
	Ret->SliceNo	= 0;

	Ret->TreeFN		= NULL;
	Ret->Name		= StrMake(Name);
	Ret->Taxa		= NULL;

	return Ret;
}

void	FreeSTreeTaxa(STREE *STree)
{
	int Index;

	if(STree->Taxa == NULL)
		return;
	
	for(Index=0;Index<STree->NoTaxa;Index++)
	{
		free(STree->Taxa[Index]->Name);
		free(STree->Taxa[Index]);
	}

	free(STree->Taxa);
}

void	FreeSTree(STREE *STree)
{
	FreeSTreeTaxa(STree);

	if(STree->Root != NULL)
		RecFreeTree(STree->Root);
	
	if(STree->NodeList != NULL)
		free(STree->NodeList);

	if(STree->Name != NULL)
		free(STree->Name);

	if(STree->TreeFN != NULL)
		free(STree->TreeFN);
	free(STree);
}

void	AddSTreeNode(STREE *STree, NODE Node, NODE Ans)
{
	if(Ans == NULL)
	{
		STree->Root = Node;
		return;
	}

	Ans->NList = (NODE*)AddToList(&Ans->NoNodes, (void**)Ans->NList, (void*) Node);
	Node->Ans = Ans;
}

void	BuildSTree(STREE *STree, NODE Ans, NODE Node, double Cat, double DistSF)
{
	int Index;
	NODE NNode;

	if((Cat > DistSF + Node->TLen) && (Node->Tip == TRUE))
		return;

	NNode = CloneNode(Node);
	AddSTreeNode(STree, NNode, Ans);

	
	if((Cat > DistSF) && (Cat < DistSF + Node->TLen))
	{
		NNode->Tip = TRUE;
		NNode->Len = Cat - DistSF;
		NNode->TLen = Cat - DistSF;

		return;
	}
	
	for(Index=0;Index<Node->NoNodes;Index++)
		BuildSTree(STree, NNode, Node->NList[Index], Cat, DistSF+Node->TLen);
}

void GetNoOfNode(NODE Node, int *No)
{
	int	Index;

	(*No)++;
	for(Index=0;Index<Node->NoNodes;Index++)
		GetNoOfNode(Node->NList[Index], No);
}

void	FlattenSNodeList(NODE Node, NODE* List, int *No)
{
	int	Index;

	List[*No] = Node;
	(*No)++;
	for(Index=0;Index<Node->NoNodes;Index++)
		FlattenSNodeList(Node->NList[Index], List, No);

}

void	MakeSTaxa(NODE N, int No, char *Buffer)
{
	TAXA	*Ret;

	Ret = (TAXA*)malloc(sizeof(TAXA));
	if(Ret == NULL)
		MallocErr();

	N->Taxa = Ret;
	Ret->No = No;

	sprintf(Buffer, "Taxa-%06d", No);
	Ret->Name = StrMake(Buffer);
}

void	MakeSTreeTaxa(STREE *STree)
{
	int		Index;
	char	*Buffer;
	NODE	Node;

	Buffer = (char*)malloc(sizeof(char) * BUFFERSIZE);
	if(Buffer == NULL)
		MallocErr();

	STree->NoTaxa = 0;
	for(Index=0;Index<STree->NoNodes;Index++)
		if(STree->NodeList[Index]->Tip == TRUE)
			STree->NoTaxa++;

	STree->Taxa = (TAXA**)malloc(sizeof(TAXA*) * STree->NoTaxa);
	if(STree->Taxa == NULL)
		MallocErr();

	STree->NoTaxa = 0;
	for(Index=0;Index<STree->NoNodes;Index++)
	{
		Node = STree->NodeList[Index];
		if(Node->Tip == TRUE)
		{
			MakeSTaxa(Node, Node->Pos, Buffer);
			STree->Taxa[STree->NoTaxa] = Node->Taxa;
			STree->NoTaxa++;
		}
	}

	free(Buffer);
}

void	WriteSNode(FILE *Out, NODE N, DATA *Data)
{
	int		Pos;
	BLINFO	*BLInfo;
	int		Index;

	BLInfo = Data->BLInfo;

	Pos = N->Pos;

	if(N->Tip == TRUE)
	{
		fprintf(Out, "%d:%f", N->Taxa->No, N->Len);
		return;
	}

	if(N->NoNodes == 0)
		return;

	fprintf(Out, "(");

	for(Index=0;Index<N->NoNodes-1;Index++)
	{
		WriteSNode(Out, N->NList[Index], Data);
		fprintf(Out, ",");
	}
	WriteSNode(Out, N->NList[Index], Data);

	if(N->Ans != NULL)
	{
		if(Pos != -1)
			fprintf(Out, "):%f", N->Len);
		else
			fprintf(Out, "):0.000");
	}
	else
		fprintf(Out, ")");
}

void	SaveTrees(DATA *Data, STREE *STree, char *Dir)
{
	char	*Buffer;
	FILE	*Out;
	int		Index;
	TAXA	*Taxa;

	Buffer = (char*)malloc(sizeof(char) * BUFFERSIZE);
	if(Buffer == NULL)
		MallocErr();

	sprintf(Buffer, "%s\\%s-%06d.trees", Dir, STree->Name, STree->SliceNo);

	STree->TreeFN = StrMake(Buffer);
	Out = OpenWrite(Buffer);
	free(Buffer);

	fprintf(Out, "#Nexus\n\n");
	fprintf(Out, "Begin trees;\n");
	fprintf(Out, "\tTranslate\n");

	for(Index=0;Index<STree->NoTaxa-1;Index++)
	{
		Taxa = STree->Taxa[Index];
		fprintf(Out, "\t\t%d\t%s,\n", Taxa->No, Taxa->Name);
	}
	Taxa = STree->Taxa[Index];
	fprintf(Out, "\t\t%d\t%s;\n", Taxa->No, Taxa->Name);

	fprintf(Out, "\t\ttree %s = ", STree->Name);
	WriteSNode(Out, STree->Root, Data);	

	fprintf(Out, ";\nend;\n");
	
	fclose(Out);
}

NODE	MakeNewResNode(int NoNodes, NODE* List)
{
	NODE	Ret;
	int		Index;

	Ret = (NODE)malloc(sizeof(struct INODE));
	if(Ret == NULL)
		MallocErr();

	Ret->NoNodes = NoNodes;
	Ret->NList = (NODE*)malloc(sizeof(NODE) * NoNodes);
	if(Ret->NList == NULL)
		MallocErr();

	for(Index=0;Index<NoNodes;Index++)
	{
		List[Index]->Ans = Ret;
		Ret->NList[Index] = List[Index];
	}

	Ret->Len	= 0;
	Ret->Scale	= 0;
	Ret->ID		= -1;
	Ret->Pos	= -1;
	Ret->Tip	= FALSE;

	return Ret;
}

void	ResloveTree(STREE *STree, NODE N)
{
	NODE NewN;
	NODE *NList;
	int	Index;

	if(N->NoNodes > 2)
	{
		NewN = 	MakeNewResNode(N->NoNodes-1, &N->NList[1]);
		NewN->Ans = N;

		NList = (NODE*)malloc(sizeof(NODE) * 2);
		if(NList == NULL)
			MallocErr();
		NList[0] = N->NList[0];
		NList[1] = NewN;
		free(N->NList);

		N->NList = NList;
		N->NoNodes = 2;
	}

	if(N->Tip == TRUE)
		return;

	for(Index=0;Index<N->NoNodes;Index++)
		ResloveTree(STree, N->NList[Index]);
}

void	SetTreeBL(NODE N)
{
	int Index;

#ifdef PHYCORRUM
	N->Len = N->TLen;
#else
	N->Len = N->TLen * N->Scale;
#endif

	for(Index=0;Index<N->NoNodes;Index++)
		SetTreeBL(N->NList[Index]);
}

void	RepAns(NODE Ans, NODE Rep, NODE With)
{
	int Index;
	
	for(Index=0;Index<Ans->NoNodes;Index++)
	{
		if(Ans->NList[Index] == Rep)
		{
			Ans->NList[Index] = With;
			return;
		}
	}
}


void	DelNode(NODE N, NODE Del)
{
	NODE *NList;
	int  Index, I;


	if(N->NoNodes == 1)
	{
		free(N->NList);
		N->NList = NULL;
		N->NoNodes = 0;
		return;
	}

	NList = (NODE*)malloc(sizeof(NODE) * (N->NoNodes - 1));
	if(NList == NULL) 
		MallocErr();

	I = 0;
	for(Index=0;Index<N->NoNodes;Index++)
		if(N->NList[Index] != Del)
			NList[I++] = N->NList[Index];

	free(N->NList);
	N->NList = NList;
	N->NoNodes--;
}

void	CheckNode(NODE N, int *Err)
{
	int	Index;

	if(N->Tip == TRUE)
	{
		printf("Tip\t%f\t|\t", N->Len);
		return;
	}
	

	if(N->NoNodes < 2)
	{
		Index = N->NoNodes;
		printf("Node Err.\n");
	} 

	printf("Node\t%d\t%f\t|\t", N->NoNodes, N->Len);
		
	if(N->NoNodes < 2)
	{
		*Err = TRUE;
//		return;
	}

	for(Index=0;Index<N->NoNodes;Index++)
		CheckNode(N->NList[Index], Err);
}

int	CheckTree(STREE *Tree)
{
	int Err;

	Err = FALSE;

	if(Tree->Root == NULL)
		return Err;

	if(Tree->Root->Tip == TRUE)
		return Err;

	CheckNode(Tree->Root, &Err);

	return Err;
}

int		CountTips(NODE N)
{
	int Tips, Index;

	if(N->Tip == TRUE)
		return 1;
	
	Tips = 0;
	for(Index=0;Index<N->NoNodes;Index++)
		Tips += CountTips(N->NList[Index]);
	
	return Tips;
}

void	OutputDummyData(DATA *Data, STREE *STree, char *Dir)
{
	char	*Buffer;
	FILE	*Out;
	
#ifndef SAVEFILES
	return;
#endif

//	printf("Its me.\n");exit(0);

	Buffer = (char*)malloc(sizeof(char) * BUFFERSIZE);
	if(Buffer == NULL)	MallocErr();

	sprintf(Buffer, "%s\\%s-%06d.trees.txt", Dir, STree->Name, STree->SliceNo);
	Out = OpenWrite(Buffer);

	if(STree->Root != NULL)
		fprintf(Out, "Err with Slice %d\tN\\A\t%f\tN\\A\n", STree->SliceNo, Data->BLInfo->MeanScale[STree->Root->Pos]);
	else
		fprintf(Out, "Err with Slice %d\tN\\A\tN\\A\tN\\A\n", STree->SliceNo);

	
	fclose(Out);
	free(Buffer);
}



void	RecCalcContrast(NODE N)
{
	int Index;
	NODE	Node0, Node1;
	double	l0, l1, t;

	if(N->Tip == TRUE)
		return;

	for(Index=0;Index<N->NoNodes;Index++)
		RecCalcContrast(N->NList[Index]);

	Node0 = N->NList[0];
	Node1 = N->NList[1];

	l0 = Node0->Len + Node0->Err;
	l1 = Node1->Len + Node1->Err;

	t = (l0 * Node1->Data) + (l1 * Node0->Data);
	t = t / (l0 + l1);
	
	N->Data = t;
	N->Cont = Node0->Data - Node1->Data;

	N->Err = (l0 * l1) / (l0 + l1);
	N->Var = l0 + l1;

}

void	SetConData(DATA *Data, STREE *STree)
{
	int Index;
	NODE N;

	for(Index=0;Index<STree->NoNodes;Index++)
	{
		N = STree->NodeList[Index];
		N->Data = N->Err = N->Cont = N->Var = 0;
		if(N->Tip == TRUE)
			N->Data = N->Scale;
		
	}
}

void	CalcInConParam(STREE *STree, double *Alpha, double *Sigma, double *Lh)
{
	int		Index, NoCon;
	NODE	N;
	double	GlobalVar, SumLogVar, T1;

	NoCon = 0;
	GlobalVar = 0;
	SumLogVar = 0;

	for(Index=0;Index<STree->NoNodes;Index++)
	{
		N = STree->NodeList[Index];
		if(N->Tip == FALSE)
		{
			GlobalVar += (N->Cont * N->Cont) / N->Var;
			SumLogVar += log(N->Var);
			NoCon++;
		}
	}

	SumLogVar += log(STree->Root->Err);

	T1 = GlobalVar;
	GlobalVar = GlobalVar / (NoCon + 1);
	*Lh = (NoCon + 1) * log(6.28318530717958647692528676655900576839 * GlobalVar);
	*Lh += SumLogVar + (T1 / GlobalVar);
	*Lh *= -0.5;

	*Alpha = STree->Root->Data;
	*Sigma = GlobalVar;
}

void	CalcIndepContrast(DATA *Data, STREE *STree)
{
	double Alpha, Sigma, Lh;

	SetConData(Data, STree);

	RecCalcContrast(STree->Root);

	CalcInConParam(STree, &Alpha, &Sigma, &Lh);

	printf("Yes\t%f\t%f\t%f\n", Lh, Alpha, Sigma);

}

void ProcessSTree(DATA *Data, STREE *STree, double Height, char *Dir)
{
	if(STree->Root == NULL)
	{
#ifdef SAVEFILES
		OutputDummyData(Data, STree, Dir);
#endif
		return;
	}

	printf("%s\t%d\t%f\t%d\t", STree->Name, STree->SliceNo, Height, CountTips(STree->Root));

	if(STree->Root->Tip == TRUE)
	{
		printf("No\t0\t%f\t0\n", STree->Root->Scale);
#ifdef SAVEFILES
		OutputDummyData(Data, STree, Dir);
#endif
		return;
	}

	ResloveTree(STree, STree->Root);

	STree->NoNodes = 0;
	GetNoOfNode(STree->Root, &STree->NoNodes);
	
	STree->NodeList = (NODE*)malloc(sizeof(NODE*) * STree->NoNodes);
	if(STree->NodeList == NULL)
		MallocErr();

	STree->NoNodes = 0;
	FlattenSNodeList(STree->Root, STree->NodeList, &STree->NoNodes);

	MakeSTreeTaxa(STree);

	if(STree->NoNodes > 1)
	{
#ifdef SAVEFILES
		SaveTrees(Data, STree, Dir);
#endif
		CalcIndepContrast(Data, STree);
		return;
	}
	
	printf("Error with slice %d, not enough nodes\n", STree->SliceNo);

#ifdef SAVEFILES
	OutputDummyData(Data, STree, Dir);
#endif
}

void	MakeDataFile(DATA *Data, char *Dir)
{
	FILE	*Out;
	char	*Buffer;
	BLINFO	*BLInfo;
	int		Index, Pos;
	NODE	N;
	
	Buffer = (char*)malloc(sizeof(char) * BUFFERSIZE);
	if(Buffer == NULL)
		MallocErr();

	sprintf(Buffer, "%s\\Data.txt", Dir);

	Out = OpenWrite(Buffer);
	
	free(Buffer);

	BLInfo = Data->BLInfo;

	for(Index=0;Index<Data->NoNodes;Index++)
	{
		N = Data->NodeList[Index];
		Pos = N->Pos;
//		fprintf(Out, "Taxa-%06d\t%f\n", N->Pos, BLInfo->MeanScale[Pos]);
		fprintf(Out, "Taxa-%06d\t%f\n", N->Pos, N->Scale);
	}

	fclose(Out);
}

void	GenScrips(DATA *Data)
{
	int RNo;
	RATES *Rates;
	OPTIONS *Opt;
	
	
	Opt = Data->Opt;
	for(RNo=0;RNo<Opt->NoRates;RNo++)
	{
		Rates = Opt->Rates[RNo];
		printf("nohup nice -19 find -name \"%s-*.trees\" -printf \"./BayesTraits.exe %%p Data.txt < in.txt > %%p.txt\\n\" | sort  | sh &\n", Rates->Name);
	}
	printf("wait\n");

	for(RNo=0;RNo<Opt->NoRates;RNo++)
	{
		Rates = Opt->Rates[RNo];
		printf("nohup nice -19 find -name \"%s*.trees.txt\" -printf \"tail -n 1 %%p\\n\" | sort | sh > Res-%s.txt &\n", Rates->Name, Rates->Name);
	}
	printf("wait\n");

	for(RNo=0;RNo<Opt->NoRates;RNo++)
	{
		Rates = Opt->Rates[RNo];
		printf("find -name \"%s*.trees.txt\" -printf \"rm %%p\\n\" | sort | sh\n", Rates->Name);
	}
}

void	SetRoot(STREE *STree)
{
	NODE TRoot;

	if(STree->Root->NoNodes > 1)
		return;

	if(STree->Root->Tip == TRUE)
		return;

	if(STree->Root->NoNodes == 0)
	{
		free(STree->Root);
		STree->Root = NULL;
		return;
	}

	while(STree->Root->NoNodes < 2)
	{
		TRoot = STree->Root;
		if(TRoot->Tip == TRUE)
			return;

		if(TRoot->NoNodes == 0)
		{
			free(STree->Root);
			STree->Root = NULL;
			return;
		}
		
		STree->Root = TRoot->NList[0];
		free(TRoot->NList);
		free(TRoot);
	}
}

int		VTree(NODE *NList, int Len)
{
	int Index;
	NODE N;

	for(Index=1;Index<Len;Index++)
	{
		N = NList[Index];
		if(N != NULL)
		{
			if(N->Tip == FALSE)
			{
				if(N->NoNodes < 2)
					return FALSE;
			}
		}
	}

	return TRUE;
}

int		ValidRoot(NODE N)
{
	if(N->Tip == TRUE)
		return FALSE;

	if(N->NoNodes == 0)
		return FALSE;

	return TRUE;
}

NODE	NormNode(NODE N)
{
	NODE TNode;
	if(N == NULL)
		return N;

	if(N->Ans == NULL)
		return N ;

	if(N->Tip == TRUE)
		return N;

	if(N->NoNodes == 0)
	{
		DelNode(N->Ans, N);		
		free(N);
		return NULL;
	}

	if(N->NoNodes == 1)
	{
		TNode = N->NList[0];
		RepAns(N->Ans, N, TNode);
		TNode->Len += N->Len;
		TNode->Ans = N->Ans;
		free(N->NList);
		free(N);
		return NULL;
	}
	
	return N;
}

void	PruneTree(STREE *STree)
{
	NODE	*NList;
	int		Len, Index;

	Len = 0;
	GetNoOfNode(STree->Root, &Len);

	NList = (NODE*)malloc(sizeof(NODE) * Len);
	if(NList == NULL) MallocErr();
	
	Len = 0;
	FlattenSNodeList(STree->Root, NList, &Len);

	while(VTree(NList, Len) == FALSE)
	{
		if(ValidRoot(STree->Root) == FALSE)
		{
			free(NList);
			return;
		}

		for(Index=0;Index<Len;Index++)
			NList[Index] = NormNode(NList[Index]);
	}

	free(NList);
}


int		ValidTree(STREE *STree)
{
	PruneTree(STree);
	SetRoot(STree);

//	CheckTree(STree);
	return TRUE;
}

void	PhyCorrectCon(DATA *Data, char *Dir)
{
	OPTIONS		*Opt;
	int			RNo, CIndex;
	RATES		*Rates;
	STREE		*STree;

#ifdef SAVEFILES
	mdir(Dir);
	MakeDataFile(Data, Dir);
#endif

	SetConBLScale(Data);

	Opt = Data->Opt;

	for(RNo=0;RNo<Opt->NoRates;RNo++)
	{
		Rates = Opt->Rates[RNo];
		printf("Phylogenetically Corrected Rates For %s\n", Rates->Name);
		printf("Rates\tSlice\tSlice Height\tTips\tValid\tLh\tAlpha\tSigma\n");

		for(CIndex=0;CIndex<Rates->NoCats;CIndex++)
		{ 
			if(Rates->Cats[CIndex] >= Rates->StatPos)
			{
				STree = MakeSTree(Rates->Name);
				STree->SliceNo = CIndex;
				BuildSTree(STree, NULL, Rates->StartNode, Rates->Cats[CIndex], Rates->StatPos);

				SetTreeBL(STree->Root);
				ValidTree(STree);
				
				ProcessSTree(Data, STree, Rates->Cats[CIndex], Dir);
				FreeSTree(STree);
			}
			else
				printf("Slice\t%s\tNode at slice has not started yet\n", Rates->Name);
		} 
	} 

#ifdef SAVEFILES
	printf("Commands to run.\n");
	GenScrips(Data);
#endif
}
